#!/usr/bin/env python

# placeholder for the script that downloads metadata from glance and
# places them in the juju-configured data_dir


import os
import sys
import argparse
import json
import datetime
import dateutil.parser
import atexit
import os_client_config


class BootStackMetadataError(Exception):
    """Bootstack chosen extra metadata keys now in use
        * bs_owner:
            name of the tenant an image is assigned to
            helps map different tenants between regions (with
           different tenant-id, but same/similar tenant-name
    """
    pass


class ImageSyncMaster:
    def __init__(self, data_dir='/srv/glance_master_sync/data'):
        self.tenants = {}
        self.DATA_DIR = data_dir
        TMP_DIR = '/tmp/glance_master_sync'
        if not os.path.isdir(TMP_DIR):
            os.makedirs(TMP_DIR)

        self.set_filelock()
        self.glance_connect()

    def glance_connect(self):
        try:

            self.log('connecting to Keystone')
            self.keystone = os_client_config.session_client(
                    'identity',
                    cloud='envvars')
        except Exception as e:
            self.log('EXCEPTION: {0}'.format(e))
            sys.exit(2)
        if not self.tenants:
            self.tenants = dict(
                    [(tenant['id'], tenant['name']) for tenant in
                     self.keystone.get('/projects').json()['projects'] if
                     tenant['enabled']])
        self.glance = os_client_config.session_client('image', cloud='envvars')
        return self.glance

    def timestamp_now(self):
        return datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")

    def log(self, msg):
        print('{0} {1}'.format(self.timestamp_now(), msg))

    def delete_files(self, existing_images_ids):
        if not existing_images_ids:
            self.log('WARNING: precautionary halt. '
                     'No glance images found. noop.')
            return

        for dirpath, dirnames, filenames in os.walk(self.DATA_DIR):
            if dirpath != self.DATA_DIR and \
               len(dirnames) == 0 and \
               len(filenames) == 0:
                os.rmdir(dirpath)
                continue

            for filename in filenames:
                full_path = os.path.join(dirpath, filename)
                if filename.endswith('.json.tmp'):
                    self.log('WARNING: temporary file skipped. Please check '
                             '{0}'.format(full_path))
                    continue
                elif filename.endswith('.json') and \
                        filename[:-5] in existing_images_ids:
                    continue
                else:
                    self.log('INFO: image not found in glance - deleting '
                             '{0}'.format(full_path))
                    os.remove(full_path)

    def create_lock(self, lockfile):
        try:
            os.open(lockfile, os.O_CREAT | os.O_EXCL | os.O_RDWR)
        except OSError:
            self.log('ERROR: could not create lockfile {}'.format(lockfile))

    def file_locked(self, lockfile='/tmp/glance_sync_master.lock'):
        if os.path.isfile(lockfile):
            return True
        else:
            return False

    def release_lock(self, lockfile='/tmp/glance_sync_master.lock'):
        if os.path.isfile(lockfile):
            try:
                os.remove(lockfile)
            except OSError as e:
                self.log(e)

    def set_filelock(self, lockfile='/tmp/glance_sync_master.lock'):
        if self.file_locked(lockfile):
            self.log('WARNING: sync already in progress, exiting')
            sys.exit(2)

        self.create_lock(lockfile)
        atexit.register(self.release_lock)

    def parse_glance_images(self):
        """
        image.id => [0-9a-z]*
            len(image.id) < 2 : DATA_DIR/<image-id>.json
            len(image.id) >= 2: DATA_DIR/XX/XXZZZ.json
        """
        self.log('getting image list')
        existing_images = set()
        for image in self.glance.get('/v2/images').json()['images']:
            existing_images.add(image['id'])
            if len(image['id']) < 2:
                basename = '{0}.json'.format(image['id'])
            else:
                basename = '{0}/{1}.json'.format(str(image['id'])[:2],
                                                 image['id'])
            filename = os.path.join(self.DATA_DIR, basename)
            if not self.is_latest_metadata(filename, image['updated_at']):
                self.update_metadata(filename, image)

        return existing_images

    def is_latest_metadata(self, filename, glance_updated_at):
        """Compares filename content (JSON metadata) and glance service info
        @return
            True: no need to update
            False: need to update metadata on local copy
        """
        if os.path.isfile(filename):
            with open(filename) as meta_file:
                try:
                    data = json.load(meta_file)
                    local_updated_at = data['updated_at']
                    imageid = data['id']
                except Exception as e:
                    self.log('EXCEPTION: {0}'.format(e))
                    return False

            local_dup = dateutil.parser.parse(local_updated_at)
            glance_dup = dateutil.parser.parse(glance_updated_at)

            if local_dup >= glance_dup:
                self.log('INFO: {0} up to date'.format(imageid))
                return True
            else:
                self.log('INFO: {0} outdated. Re-creating local '
                         'copy of the metadata'.format(imageid))
        else:
            self.log('INFO: {0} not found. Creating a local '
                     'copy'.format(filename))
            return False

    def update_metadata(self, filename, glance_metadata):
        """creates or replaces file with image metadata
           creates subdirectory if it doesn't exist
        """
        tmp_file = '{0}.tmp'.format(filename)

        if not os.path.exists(os.path.dirname(filename)):
            os.mkdir(os.path.dirname(filename), 0750)

        try:
            glance_metadata = self.add_bs_metadata_keys(glance_metadata)
        except BootStackMetadataError, e:
            self.log(e)
            return False

        with open(tmp_file, 'w') as f:
            dump = json.dumps(glance_metadata, indent=4, ensure_ascii=False)
            f.write(dump.encode('utf8'))
        os.rename(tmp_file, filename)
        self.log('INFO: update_metadata :: {0}'.format(filename))
        return True

    def add_bs_metadata_keys(self, glance_metadata):
        keys = [k for k in glance_metadata if k.startswith('bs_')]
        if 'bs_owner' in keys:
            msg = 'WARNING: bs_owner metadata should not exist (image: ' \
                  '{0}; bs_owner: {1})'.format(glance_metadata.id,
                                               glance_metadata.bs_owner)
            raise BootStackMetadataError(msg)
        elif glance_metadata['owner'] in self.tenants:
            glance_metadata['bs_owner'] = self.tenants[glance_metadata['owner']]
        return glance_metadata

    def main(self):
        self.log('starting glance sync')
        # updates local metadata files if outdated
        existing_images = self.parse_glance_images()
        # removes local metadata files from deleted images
        self.delete_files(existing_images)
        self.log('ending glance sync')
        self.release_lock()


if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='Synchronize glance images '
                                                 'to disk ')
    parser.add_argument("-d", "--datadir", help="directory to write images to")
    args = parser.parse_args()

    if args.datadir:
        data_dir = args.datadir
    else:
        parser.print_help()
        sys.exit('ERROR: please specify an output directory for images')

    master = ImageSyncMaster(data_dir)
    master.main()
