#!/usr/bin/env python3

import unittest
import os
import json
import uuid
import hashlib
import subprocess
import signal
import shlex

from glance_sync_master import ImageSyncMaster


# This script tests the glance image sync script.
# It is meant to run during user acceptance tests and will download images
# and metadata from glance and copy them into the specified data_dir from where
# they can get pull-synchronized from a slave via rsync over ssh.
#
# Since glance does not fully remove images from its database
# on image deletion you will need to use the database cleanup script
# after user acceptances tests have been completed.
# It's located in the script_dir (default: /srv/glance_sync_master/scripts) and
# will have to be copied to a glance unit to be run.

class TestMasterSync(unittest.TestCase):

    def setUp(self):
        self.data_dir = '/srv/glance_sync_master/data/'
        if not os.path.exists(self.data_dir):
            os.makedirs(self.data_dir)

        self.temp_dir = '/tmp/canonical-glance-sync-tests/'
        if not os.path.exists(self.temp_dir):
            os.makedirs(self.temp_dir)

        self.lockfile = '/tmp/glance_sync_master.lock'
        self.master = ImageSyncMaster()

        self.glance = self.master.glance_connect()
        self.lockfile = '/tmp/glance_sync_slave.lock'

    def test_sync_can_connect_to_glance(self):
        """ Test if sync script can connect to glance """
        glance = self.glance
        glance.images.list()

    def test_image_download(self):
        """ Test if sync script can download images from glance """

        # create test image
        image_id = self.create_image_in_glance(self.temp_dir)
        print(image_id)

        # get image metadata from glance
        image = self.glance.images.get(image_id)

        # download the image to disk
        self.master.main(self.data_dir)

        # get metadata from downloaded json file
        downloaded_image = os.path.join(self.data_dir, image_id)
        downloaded_metafile = os.path.join(self.data_dir, image_id + '.json')
        with open(downloaded_metafile, 'r') as meta_file:
            downloaded_metadata = json.load(meta_file)

        # verify image checksum matches metadata
        self.assertTrue(image['name'] == downloaded_metadata['name'])

        # cleanup
        self.glance.images.delete(image_id)
        os.remove(downloaded_image)
        os.remove(downloaded_metafile)

    def test_metadata_download(self):
        """ Test if sync script can download metadata from glance """

        # create test image
        image_id = self.create_image_in_glance(self.temp_dir)
        print(image_id)

        # get metadata from glance
        image = self.glance.images.get(image_id)

        # download image and metadata
        self.master.main(self.data_dir)

        # verify all fields in glance are represented in the json file
        downloaded_metafile = os.path.join(self.data_dir, image_id + '.json')
        with open(downloaded_metafile, 'r') as meta_file:
            downloaded_metadata = json.load(meta_file)

        for key in image.keys():
            if downloaded_metadata[key] != image[key]:
                self.fail('metadata mismatch: {}'.format(key))

        # cleanup
        self.glance.images.delete(image_id)
        self.master.main(self.data_dir)

    def test_metadata_update(self):
        ("Test if metadata updates in glance are reflected on filesystem")
        # have an image in glance with matching file and metadata on disk
        image_id = self.create_image_in_glance(self.data_dir)
        print(image_id)
        self.master.main(self.data_dir)

        # change the metadata in glance
        metadata_file = os.path.join(self.data_dir, image_id + '.json')
        metadata = self.read_metadata(metadata_file)
        newname = "canonical-user-acceptance-test-metadata-changed"
        metadata['name'] = newname
        self.glance.images.update(metadata['id'], **metadata)

        # trigger a sync
        self.master.main(self.data_dir)

        # verify the metadata property has changed on disk
        synced_metadata = self.read_metadata(metadata_file)
        self.assertTrue(metadata['name'] == synced_metadata['name'])

        # cleanup
        self.glance.images.delete(image_id)
        self.master.main(self.data_dir)

    def test_sync_deletes_removed_files(self):
        ("Test if sync removes files from sync directory if glance image has "
         "been deleted")

        # have an image in glance
        image_id = self.create_image_in_glance(self.data_dir)
        print(image_id)

        # trigger a sync
        self.master.main(self.data_dir)

        # verify image and metadata exist on disk
        image_file = os.path.join(self.data_dir, image_id)
        metadata_file = os.path.join(self.data_dir, image_id + '.json')
        self.assertTrue(os.path.exists(image_file))
        self.assertTrue(os.path.exists(metadata_file))

        # delete the image from glance
        self.glance.images.delete(image_id)

        # trigger a sync
        self.master.main(self.data_dir)

        # verify image and metadata are gone from disk
        self.assertTrue(not os.path.exists(image_file))
        self.assertTrue(not os.path.exists(metadata_file))

        # no cleanup necessary

    def test_sync_only_updates_metadata_on_existing_images(self):
        ("Test if sync script avoids re-download of existing images in sync "
         "directory but updates metadata")
        # have an image in glance
        image_id = self.create_image_in_glance(self.data_dir)
        print(image_id)
        image_file = os.path.join(self.data_dir, image_id)
        meta_file = os.path.join(self.data_dir, image_id + '.json')

        # trigger a sync
        self.master.main(self.data_dir)
        first_image_time = os.path.getmtime(image_file)
        first_metadata_time = os.path.getmtime(meta_file)

        # don't change anything

        # sync again
        self.master.main(self.data_dir)

        # verify mtime on image doesn't change (metadata gets written
        # every time)
        second_image_time = os.path.getmtime(image_file)
        second_metadata_time = os.path.getmtime(meta_file)

        self.assertTrue(first_image_time == second_image_time)
        self.assertTrue(first_metadata_time != second_metadata_time)

        # cleanup
        self.glance.images.delete(image_id)
        self.master.main(self.data_dir)

    def test_sync_respects_lockfile(self):
        """ Test if sync script respects the lockfile of another sync run """

        # make sure we're good to go
        self.assertFalse(self.master.file_locked(self.lockfile))

        # start a sync in the background, copy from data_dir to data_dir just
        # to keep it busy for a bit
        command = "python glance_sync_master.py -d {}".format(self.data_dir)

        with open(os.devnull, "w") as DEVNULL:
            blocker = subprocess.Popen(shlex.split(command),
                                       stdout=DEVNULL,
                                       stderr=DEVNULL)

        # suspend that process
        blocker.send_signal(signal.SIGSTOP)

        # start a conflicting sync run
        # verify the second sync is aborted with a message
        try:
            subprocess.check_output(['python', 'glance_sync_master.py',
                                     '-d', self.data_dir])
        except subprocess.CalledProcessError as e:
            print(e.output)
            self.assertTrue("sync already in progress" in e.output)

        # let the blocker finish
        blocker.send_signal(signal.SIGCONT)
        blocker.wait()

    def check_md5(self, control_md5, image_file):
        return self.get_md5(image_file) == control_md5

    def get_md5(self, image_file):
        hash_md5 = hashlib.md5()
        with open(image_file, "rb") as f:
                for chunk in iter(lambda: f.read(4096), b""):
                    hash_md5.update(chunk)
        return hash_md5.hexdigest()

    def read_metadata(self, metadata_file):
        with open(metadata_file) as meta_file:
            data = json.load(meta_file)
        return data

    def create_image_on_disk(self, data_dir, image_id):

        # generate a random iso and the corresponding json metadata file
        image_file = os.path.join(data_dir, image_id)
        with open(image_file, 'wb') as test_image:
            test_image.write(os.urandom(1024))

        checksum = self.get_md5(image_file)
        test_metadata = {'id': image_id,
                         'checksum': checksum,
                         'name': 'canonical-user-acceptance-test-random-image',
                         'disk_format': 'qcow2',
                         'container_format': 'bare'}

        meta_file = os.path.join(data_dir,
                                 "{}.json".format(image_id))
        with open(meta_file, 'w') as dumpfile:
            json.dump(test_metadata, dumpfile)

    def create_image_in_glance(self, directory):
        # create an image in glance
        image_id = str(uuid.uuid4())
        print(image_id)

        temp_image = os.path.join(directory, image_id)

        temp_metadata = {'id': image_id,
                         'name': 'canonical-user-acceptance-test-random-image',
                         'disk_format': 'qcow2',
                         'container_format': 'bare'}

        # generate a random iso and the corresponding json metadata file
        with open(temp_image, 'wb') as test_image:
            test_image.write(os.urandom(1024))

        temp_metadata = {'id': image_id,
                         'name': 'canonical-user-acceptance-test-random-image',
                         'disk_format': 'qcow2',
                         'container_format': 'bare'}

        self.glance.images.create(**temp_metadata)
        self.glance.images.upload(temp_metadata['id'],
                                  open(temp_image))
        os.remove(temp_image)
        return image_id
