#!/usr/bin/env python

import os
import pwd
import grp
import base64
import shutil
from charmhelpers.core import hookenv, templating
from charmhelpers.core.hookenv import open_port
from charmhelpers import fetch
from charms.reactive import hook, when, set_state, remove_state
import re
import subprocess


@hook('install')
def install_glance_sync_master():
    hookenv.status_set('maintenance', 'Installing')
    fetch.apt_update()
    packages = ['python-openstackclient', 'python-nose', 'python-dateutil']
    fetch.apt_install(packages)

    # configure the variables that have a default value
    configure_config_dir()
    configure_data_dir()
    configure_log_dir()
    configure_script_dir()

    install_sync_script()
    install_db_cleanup_script()
    configure_novarc()

    homedir = os.path.expanduser('~ubuntu')
    ssh_identity = '{}/.ssh/id_rsa'.format(homedir)

    if not os.path.exists(ssh_identity):
        command = ['ssh-keygen', '-t', 'rsa', '-N', '', '-f', ssh_identity]
        proc = subprocess.Popen(command,
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE)
        (stdout, stderr) = proc.communicate()
        if proc.returncode:
            print("ERROR: problem generating ssh key '{}':".format(command))
            print(stderr)
        os.chown(ssh_identity,
                 pwd.getpwnam('ubuntu').pw_uid,
                 grp.getgrnam('ubuntu').gr_gid)

    open_port(22, protocol='TCP')
    set_state('glance-sync-master.installed')
    hookenv.status_set('active', 'Unit ready')


@hook('upgrade-charm')
def upgrade_glance_sync_master():
    install_glance_sync_master()


@when('config.changed.config_dir')
def configure_config_dir():
    hookenv.status_set('maintenance', 'Configuring')
    config_dir = hookenv.config('config_dir')

    if not os.path.exists(config_dir):
        os.makedirs(config_dir)
        os.chown(config_dir,
                 pwd.getpwnam('ubuntu').pw_uid,
                 grp.getgrnam('ubuntu').gr_gid)
    configure_cron()
    hookenv.status_set('active', 'Unit ready')


@when('config.changed.trusted_ssl_ca')
def fix_ssl():
    config = hookenv.config()
    cert_file = '/usr/local/share/ca-certificates/openstack-ca.crt'
    trusted_ssl_ca = config.get('trusted_ssl_ca').strip()
    hookenv.log("Writing ssl ca cert:{}".format(trusted_ssl_ca))
    cert_content = base64.b64decode(trusted_ssl_ca).decode()
    with open(cert_file, 'w') as f:
        print(cert_content, file=f)
    subprocess.call(["/usr/sbin/update-ca-certificates"])


@when('config.changed.data_dir')
def configure_data_dir():
    hookenv.status_set('maintenance', 'Configuring')
    data_dir = hookenv.config('data_dir')

    if not os.path.exists(data_dir):
        os.makedirs(data_dir)
        os.chown(data_dir,
                 pwd.getpwnam('ubuntu').pw_uid,
                 grp.getgrnam('ubuntu').gr_gid)
    configure_cron()
    hookenv.status_set('active', 'Unit ready')


@when('config.changed.log_dir')
def configure_log_dir():
    hookenv.status_set('maintenance', 'Configuring')
    log_dir = hookenv.config('log_dir')
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)
        os.chown(log_dir,
                 pwd.getpwnam('ubuntu').pw_uid,
                 grp.getgrnam('ubuntu').gr_gid)

    templating.render(source='logrotate.d.j2',
                      target='/etc/logrotate.d/glance_sync_master',
                      owner='root',
                      perms=0o644,
                      context=hookenv.config())
    configure_cron()
    hookenv.status_set('active', 'Unit ready')


@when('config.changed.script_dir')
def configure_script_dir():
    hookenv.status_set('maintenance', 'Configuring')
    script_dir = hookenv.config('script_dir')
    if not os.path.exists(script_dir):
        os.makedirs(script_dir)
        os.chown(script_dir,
                 pwd.getpwnam('ubuntu').pw_uid,
                 grp.getgrnam('ubuntu').gr_gid)
    install_sync_script()
    install_db_cleanup_script()
    configure_cron()
    hookenv.status_set('active', 'Unit ready')


@when('config.changed.authorized_keys')
def configure_authorized_keys():
    hookenv.status_set('maintenance', 'Configuring')

    ssh_conf_dir = '/etc/ssh/authorized-keys'
    auth_keys_file = os.path.join(ssh_conf_dir, "glance-sync-slaves")

    authorized_keys = hookenv.config('authorized_keys')

    if authorized_keys:
        keys_bytestring = base64.b64decode(authorized_keys)
        keys = str(keys_bytestring, 'utf-8')

        if not os.path.exists(ssh_conf_dir):
            os.makedirs(ssh_conf_dir)
            os.chown(ssh_conf_dir,
                     pwd.getpwnam('root').pw_uid,
                     grp.getgrnam('root').gr_gid)

        with open(auth_keys_file, "w") as f:
            f.write(keys)
        os.chmod(auth_keys_file, 0o744)

        configure_sshd(auth_keys_file)

        hookenv.status_set('active', 'Unit ready')
    else:
        if os.path.isfile(auth_keys_file):
            os.remove(auth_keys_file)
        hookenv.status_set('active', 'Unit ready')


# Adds an AutorizedKeysFile line to /etc/ssh/sshd_config
def configure_sshd(keys_file):
    original = '/etc/ssh/sshd_config'
    temp_config = '/tmp/sshd_config'
    old_lines = []
    auth_keys_file_added = False
    auth_keys_file_existed = False
    homedirs = "%h/.ssh/authorized_keys"

    with open(original, "r") as old_file:
        for line in old_file:
            old_lines.append(line)

    with open(temp_config, "w") as new_file:
        for line in old_lines:
            if re.search(r'^AuthorizedKeysFile.*{}.*'.format(keys_file), line):
                auth_keys_file_existed = True
                new_file.write(line)
                continue
            else:
                # by default, there is no AutorizedKeysFile definition in
                # sshd_config, default behaviour is to source
                # ~/.ssh/authorized_keys (there is such a line but it's
                # commented out by default)
                # If you do specify an AuthorizedKeysFile you need to make that
                # default explicit again
                replaced = re.sub(r'^AuthorizedKeysFile\s(.*)$',
                                  r'AuthorizedKeysFile \1 {} {}'.format(
                                      homedirs, keys_file),
                                  line)
                if replaced is not line:
                    auth_keys_file_added = True
                new_file.write(replaced)

        if not auth_keys_file_existed and not auth_keys_file_added:
            new_file.write('AuthorizedKeysFile {} {}\n'.format(homedirs,
                                                               keys_file))

    shutil.copy(temp_config, original)
    os.system('sudo sshd -t && service ssh reload')


def install_sync_script():
    hookenv.status_set('maintenance', 'Installing')
    # copy sync script and user acceptance tests in place
    files = ['glance_sync_master.py',
             os.path.join('tests', 'test_glance_sync_master.py'),
             os.path.join('tests', '__init__.py')]
    script_dir = hookenv.config('script_dir')

    hookenv.log('installing master sync script')

    for file in files:
        source = os.path.join(hookenv.charm_dir(), 'files', file)
        dest = os.path.join(script_dir, file)

        if not os.path.exists(os.path.dirname(dest)):
            os.makedirs(os.path.dirname(dest))

        shutil.copy(source, dest)
        os.chown(dest,
                 pwd.getpwnam('ubuntu').pw_uid,
                 grp.getgrnam('ubuntu').gr_gid)
        os.chmod(dest, 0o640)

    hookenv.status_set('active', 'Unit ready')


def install_db_cleanup_script():
    hookenv.status_set('maintenance', 'Installing')
    # copy sync script in place
    script_dir = hookenv.config('script_dir')
    source = os.path.join(hookenv.charm_dir(),
                          'files',
                          'db_purge_deleted_glance_images.py')
    dest = os.path.join(script_dir, 'db_purge_deleted_glance_images.py')
    shutil.copy(source, dest)
    os.chown(dest,
             pwd.getpwnam('ubuntu').pw_uid,
             grp.getgrnam('ubuntu').gr_gid)
    os.chmod(dest, 0o740)
    hookenv.status_set('active', 'Unit ready')


@when('config.changed.novarc')
def configure_custom_novarc():
    configure_novarc()


@hook('{requires:keystone-admin}-relation-{joined,changed}')
def configure_relation_novarc(relation=None):
    configure_novarc()


# If we have a custom novarc config set it overrides any relation-pulled
# credentials. If the custom novarc goes away, configure_relation_novarc should
# fire.
def configure_novarc():
    hookenv.status_set('maintenance', 'Configuring novarc')

    keystone_relations = hookenv.relations_of_type('keystone-admin')
    config_dir = hookenv.config('config_dir')
    novarc_file = os.path.join(config_dir, "novarc")

    if len(hookenv.config('novarc')) > 0:
        write_custom_novarc(novarc_file)
    elif len(keystone_relations) > 0:
        write_relation_novarc(novarc_file)
    else:
        message = "Please set novarc config or add a keystone relation."
        hookenv.log("ERROR: {}".format(message))
        if os.path.isfile(novarc_file):
            os.remove(novarc_file)
            remove_state('novarc.configured')

        hookenv.status_set('blocked', message)


def write_relation_novarc(path, relation=None):
    hookenv.status_set('maintenance', 'Configuring novarc from keystone '
                                      'relation')
    hookenv.log('configuring novarc based on keystone relation')
    if not relation:
        keystone = hookenv.relations_of_type('keystone-admin')
        if len(keystone) > 0:
            relation = keystone[0]

    if 'service_password' in relation and relation['service_password']:
        templating.render(source='novarc.j2',
                          target=path,
                          owner='ubuntu',
                          perms=0o600,
                          context=relation)
        hookenv.status_set('active', 'Unit ready')
    else:
        hookenv.status_set('maintenance', 'keystone relation incomplete')


def write_custom_novarc(path):
    hookenv.status_set('maintenance', 'Configuring custom novarc')

    hookenv.log('configuring custom novarc')
    if hookenv.config('novarc'):
        with open(path, "wb") as f:
            f.write(base64.b64decode(hookenv.config('novarc')))

        os.chown(path,
                 pwd.getpwnam('ubuntu').pw_uid,
                 grp.getgrnam('ubuntu').gr_gid)
        os.chmod(path, 0o600)

        set_state('novarc.configured')
        hookenv.status_set('active', 'Unit ready')
    else:
        hookenv.status_set(
                'maintenance', 'no novarc provided (juju set '
                'glance-sync-master novarc=<base64 encoded novarc>)'
                )


@when('config.changed.sync_enabled')
@when('config.changed.cron_frequency')
def configure_cron():
    hookenv.status_set('maintenance', 'Configuring')
    sync_enabled = hookenv.config('sync_enabled')

    hookenv.log('configuring sync cronjob')
    cron_file = '/etc/cron.d/glance_sync_master'
    if sync_enabled:
        hookenv.log('adding cronjob')
        templating.render(source='glance_sync_master_cron.j2',
                          target=cron_file,
                          owner='root',
                          perms=0o640,
                          context=hookenv.config())
        set_state('cron.configured')
    else:
        hookenv.log("removing cronjob")
        if os.path.isfile(cron_file):
            os.remove(cron_file)
            remove_state('cron.configured')
    hookenv.status_set('active', 'Unit ready')


@hook('nrpe-external-master-relation-changed')
def setup_nrpe_checks(nagios):
    hookenv.status_set('maintenance', 'Configuring nrpe checks')
    config = hookenv.config()
    nagios.add_check(['/usr/lib/nagios/plugins/check_file_age',
                      '-w 14400', '-c 25200', '-f',
                     os.path.join(config['log_dir'],
                                  'glance_sync_master.log')],
                     name="check_glance_sync_master_log",
                     description="Verify age of last image sync from glance "
                                 "to disk",
                     context=config["nagios_context"],
                     unit=hookenv.local_unit())

    # copy nrpe check plugin for stale lockfiles in place
    script_dir = hookenv.config('script_dir')
    files = ['check_stale_lockfile.py']

    hookenv.log('installing stale lockfile nrpe plugin')
    for file in files:
        source = os.path.join(hookenv.charm_dir(), 'files', file)
        dest = os.path.join(script_dir, file)

        if not os.path.exists(os.path.dirname(dest)):
            os.makedirs(os.path.dirname(dest))

        shutil.copy(source, dest)
        os.chown(dest,
                 pwd.getpwnam('ubuntu').pw_uid,
                 grp.getgrnam('ubuntu').gr_gid)
        os.chmod(dest, 0o740)

    nrpe_plugin = os.path.join(script_dir, 'check_stale_lockfile.py')
    nagios.add_check([nrpe_plugin, '-f', '/tmp/glance_sync_master.lock',
                      '-w 72000', '-c 14400'],
                     name="check_glance_sync_master_lockfile",
                     description="Verify age of image sync lockfile",
                     context=config["nagios_context"],
                     unit=hookenv.local_unit())

    hookenv.status_set('active', 'Unit ready')
